@@ -2,6 +2,8 @@ import grpc
 import jwt  # pip install pyJWT, ATTENTION, no pip install jwt!!
 import time
 import os
+import logging
+from datetime import datetime
 from retry import retry
 
 from . import auth_pb2
@@ -10,6 +12,7 @@ from . import auth_pb2_grpc
 # CONSTANTS
 BX_HEADER_TOKEN_NAME = "bx_token"
 CHANNEL_TIMEOUT = 40  # In seconds
+LOGGER = logging.getLogger('batchx');
 
 # CONFIG
 CALL_DELAY = 0.5      # Initial delay (in seconds) between remote method call attempts.
@@ -54,7 +57,11 @@ def connect_to(endpoint, token, secure):
     if channel is None:
         try:
             if secure:
-                channel = grpc.secure_channel(endpoint, grpc.ssl_channel_credentials())
+                channel = grpc.secure_channel(endpoint, grpc.ssl_channel_credentials(), options={
+                    "grpc.keepalive_time_ms": 30000,
+                    'grpc.keepalive_timeout_ms': 30*60*1000,
+                    'grpc.keepalive_permit_without_calls': True
+                }.items())
             else:
                 channel = grpc.insecure_channel(endpoint)
             grpc.channel_ready_future(channel).result(CHANNEL_TIMEOUT)
@@ -97,19 +104,24 @@ def check():
 
 
 @retry(RecoverableException, delay=CALL_DELAY, backoff=CALL_BACKOFF, max_delay=CALL_MAX_DELAY, tries=MAX_TRIES)
-def make_call(func, request, include_metadata=True):
+def make_call(func, request, auth_required=True, include_timeout=True):
     """Wrapper to manage remote method calls. It implements a retry policy with exponential backoff.
 
     :param str func: Function in charge of calling remote methods.
     :param str request: grpc message.
-    :param str include_metadata: if set to True or not passed, token header will be sent.
+    :param str auth_required: if set to True or not passed, token header will be sent.
+    :param str include_timeout: if set to True or not passed, default GRPC deadline will be set.
     :return: response returned by executing the function passed as parameter (func).
     """
     try:
-        if include_metadata:
-            return func(request, metadata=[(BX_HEADER_TOKEN_NAME, access_token)], timeout=CALL_TIMEOUT)
+        if include_timeout:
+            timeout = CALL_TIMEOUT
         else:
-            return func(request, timeout=CALL_TIMEOUT)
+            timeout = None;
+        if auth_required:
+            return func(request, metadata=[(BX_HEADER_TOKEN_NAME, access_token)], timeout=timeout)
+        else:
+            return func(request, timeout=timeout)
     except grpc.RpcError as exc:
         if is_recoverable(exc):
             raise RecoverableException  # If error is recoverable, retry call
@@ -123,10 +135,10 @@ def login():
     if bx_token is not None:
         user_id = jwt.decode(bx_token, options={"verify_signature":False})['batchx-id']
         request_data = auth_pb2.RefreshTokenRequest(user_name=user_id, refresh_token=bx_token)
-        response = make_call(auth_pb2_grpc.AuthServiceStub(channel).RefreshToken, request_data, include_metadata=False)
+        response = make_call(auth_pb2_grpc.AuthServiceStub(channel).RefreshToken, request_data, auth_required=False)
         access_token = response.access_token
     else:
-        raise Exception("Batchx token not set. Please run: bx.connect(<endpoint>, <token>)")
+        raise Exception("Batchx token not set. Please run: bx.connect()")
 
 
 from . import environment_pb2
@@ -526,7 +538,7 @@ class JobService:
 
     def Stream(self, request):
         check()
-        return make_call(job_pb2_grpc.JobServiceStub(channel).Stream, request)
+        return make_call(job_pb2_grpc.JobServiceStub(channel).Stream, request, include_timeout=False)
 
     def Submit(self, request):
         check()
@@ -543,13 +555,27 @@ class JobService:
     def Run(self, submit_request):
         submit_response = self.Submit(submit_request)
         if submit_response.WhichOneof("case") == "suggested_parameters":
-            raise Exception("Unsupported job parametrization. Suggested one is: " + str(getattr(submit_response,suggested_parameters)))
-        request_stream = self.StreamRequest(environment=submit_request.environment, job_id=submit_response.job_id)
-        response = self.Stream(request_stream)
-        for r in response:
-            for job_event in r.event:
-                if job_event.WhichOneof("type") == "return_value":
-                    return getattr(job_event, job_event.WhichOneof("type"));
+            raise Exception("Unsupported job parametrization. Suggested one is: " + str(submit_response.suggested_parameters))
+        request_stream = self.StreamRequest(environment=submit_request.environment, job_id=submit_response.job_id, logs_enabled=True)
+        return self.__ProcessStream(submit_request, request_stream)
+
+    @retry(RecoverableException, delay=CALL_DELAY, backoff=CALL_BACKOFF, max_delay=CALL_MAX_DELAY, tries=MAX_TRIES)
+    def __ProcessStream(self, submit_request, request_stream):
+        try:
+            response = self.Stream(request_stream)
+            for r in response:
+                for job_event in r.event:
+                    if job_event.WhichOneof("type") == "job_status":
+                        LOGGER.warning("["+request_stream.environment+"#"+str(request_stream.job_id)+"] Job status: " + job_pb2.Status.Name(job_event.job_status))
+                    elif job_event.WhichOneof("type") == "log_record":
+                        LOGGER.warning("["+request_stream.environment+"#"+str(request_stream.job_id)+"|"+request_stream.environment+"@"+submit_request.image+"] ["+str(datetime.fromtimestamp(job_event.log_record.ts_millis/1000.0))+"] " +  job_event.log_record.message)
+                    elif job_event.WhichOneof("type") == "return_value":
+                        return job_event.return_value;
+        except grpc.RpcError as exc:
+            if is_recoverable(exc):
+                raise RecoverableException  # If error is recoverable, retry call
+            else:
+                raise exc
 
     def CancelRequest(self, **kwargs):
         return job_pb2.CancelRequest(**kwargs)
@@ -751,6 +777,10 @@ class AdminService:
         check()
         return make_call(admin_pb2_grpc.AdminServiceStub(channel).AddEnvironmentCredits, request)
 
+    def CreateMembership(self, request):
+        check()
+        return make_call(admin_pb2_grpc.AdminServiceStub(channel).CreateMembership, request)
+
     def CreateOrg(self, request):
         check()
         return make_call(admin_pb2_grpc.AdminServiceStub(channel).CreateOrg, request)
@@ -776,6 +806,12 @@ class AdminService:
     def Assignment(self, **kwargs):
         return admin_pb2.Assignment(**kwargs)
 
+    def CreateMembershipRequest(self, **kwargs):
+        return admin_pb2.CreateMembershipRequest(**kwargs)
+
+    def CreateMembershipResponse(self, **kwargs):
+        return admin_pb2.CreateMembershipResponse(**kwargs)
+
     def CreateOrgRequest(self, **kwargs):
         return admin_pb2.CreateOrgRequest(**kwargs)
 
@@ -809,14 +845,6 @@ class OrganizationService:
         check()
         return make_call(organization_pb2_grpc.OrganizationServiceStub(channel).GetOrganization, request)
 
-    def SetMembership(self, request):
-        check()
-        return make_call(organization_pb2_grpc.OrganizationServiceStub(channel).SetMembership, request)
-
-    def SetMembershipVisibility(self, request):
-        check()
-        return make_call(organization_pb2_grpc.OrganizationServiceStub(channel).SetMembershipVisibility, request)
-
     def UpdateOrganization(self, request):
         check()
         return make_call(organization_pb2_grpc.OrganizationServiceStub(channel).UpdateOrganization, request)
@@ -827,18 +855,6 @@ class OrganizationService:
     def GetOrganizationResponse(self, **kwargs):
         return organization_pb2.GetOrganizationResponse(**kwargs)
 
-    def SetMembershipRequest(self, **kwargs):
-        return organization_pb2.SetMembershipRequest(**kwargs)
-
-    def SetMembershipResponse(self, **kwargs):
-        return organization_pb2.SetMembershipResponse(**kwargs)
-
-    def SetMembershipVisibilityRequest(self, **kwargs):
-        return organization_pb2.SetMembershipVisibilityRequest(**kwargs)
-
-    def SetMembershipVisibilityResponse(self, **kwargs):
-        return organization_pb2.SetMembershipVisibilityResponse(**kwargs)
-
     def UpdateOrganizationRequest(self, **kwargs):
         return organization_pb2.UpdateOrganizationRequest(**kwargs)
 
